﻿using System;  
using System.Collections.Generic;   
using System.IO;  
using System.Net.Http;  
using System.Net.Http.Headers;  
using System.Security.Cryptography;  
using System.Text;  
using System.Threading.Tasks;  
using Newtonsoft.Json;  

// Класс для создания объектов из LINQ-запроса ADVANTA
// Примечание: LINQ-запрос GetAllObjects возвращает только Id и Name. Остальные поля надо добавить в запрос, если они нужны.
public class AdvantaObject  
{  
    public required string Id { get; set; }  
    public required string Name { get; set; }  
    public int PercentComplete { get; set; }
    public int Status { get; set; }  
    public int Priority { get; set; } 
}

// Основной класс программы
class Program  
{  
    // Адрес сервера ADVANTA
    private static string a2serverUrl = "https://pmo.a2nta.ru";

    // Логин пользователя для подключения к ADVANTA
    private static string a2login = "admin";

    // Пароль пользователя для подключения к ADVANTA
    private static string a2password = "123456";

    // Ключ LINQ-запроса в ADVANTA 
    private static string a2dataSourceKey = "GetAllObjects"; 

    // Количество объектов, получаемых из LINQ-запроса в ADVANTA (при большом количестве данные будут загружаться долго)
    private static int a2numOfObjects = 7; 

    // Имя файла с данными для авторизации в Google
    private static string gFileName = "service-account.json";
    
    // Идентификатор Google-таблицы
    private static string sheetId = "1v5lqduyizJtf9lliUkfFEjGo4j61eNX3nwV-FkZo5DU";

    // Имя листа в Google-таблице
    private static string sheetName = "Лист1";

    private static readonly HttpClient httpClient = new HttpClient(); // Общий экземпляр HttpClient для разных вызовов

    static async Task Main(string[] args)  
    {  
        try  
        {  
            // Получение токена доступа к Google
            string accessToken = await GetAccessToken(gFileName);  

            // Проверка доступа к Google-таблице  
            var range = $"{sheetName}!A1:A2";  
            if (await CheckSheetAccess(sheetId, range, accessToken))  
            {  
                Console.WriteLine("Доступ к Google-таблице успешен."); 

                // Вызов метода и получение куки ADVANTA для дальнейшего выполнения LINQ-запроса 
                string cookie = await AuthorizeInAdvanta();

                // Если авторизация успешна, выполняем следующий запрос  
                if (!string.IsNullOrEmpty(cookie))  
                {  
                    Console.WriteLine("Авторизация в ADVANTA пройдена успешно."); 

                    var advantaObjects = await GetObjectsList(cookie); // Получаем объекты из ADVANTA  

                    if (advantaObjects != null && advantaObjects.Count > 0) // Проверяем, что список не пуст  
                    {   
                        Console.WriteLine("Список объектов ADVANTA успешно получен (" + advantaObjects.Count + " шт).");

                        // Очистка таблицы  
                        await ClearSheet(sheetId, sheetName, accessToken);  

                        // Записываем новые строки в таблицу
                        await WriteAdvantaObjectsToSheet(sheetId, sheetName, advantaObjects, accessToken); // Записываем объекты в таблицу  
                    }
                }
            }  
            else  
            {  
                Console.WriteLine("Google-таблица не доступна!");  
            }  
        }  
        catch (Exception ex)  
        {  
            Console.WriteLine("Ошибка доступа: " + ex.Message);  
        }   
    }  

    // Метод для получения токена доступа на основе service-account.json  
    static async Task<string> GetAccessToken(string filePath)  
    {  
        string json = await File.ReadAllTextAsync(filePath);  
        if (string.IsNullOrWhiteSpace(json))  
        {  
            throw new Exception("Файл service-account.json пустой или не существует.");  
        }  

        var credential = JsonConvert.DeserializeObject<ServiceAccountCredential>(json);  
        if (credential == null)  
        {  
            throw new Exception("Десериализация не удалась. Убедитесь, что структура JSON правильная.");  
        }  

        // Вывод значений для отладки  
        // Console.WriteLine($"Client Email: {credential.ClientEmail}");  
        // Console.WriteLine($"Private Key: {credential.PrivateKey}");  

        if (string.IsNullOrWhiteSpace(credential.PrivateKey))  
        {  
            throw new Exception("Закрытый ключ (privateKey) не должен быть null или пустым.");  
        }  

        var token = CreateJwt(credential);  
        var accessToken = await RequestAccessToken(credential, token);  
        return accessToken;  
    }  

    // Запрос токена доступа в Google
    static async Task<string> RequestAccessToken(ServiceAccountCredential credential, string token)  
    {  
        using (var httpClient = new HttpClient())  
        {  
            var request = new HttpRequestMessage(HttpMethod.Post, "https://oauth2.googleapis.com/token");  
            request.Content = new FormUrlEncodedContent(new[]  
            {  
                new KeyValuePair<string, string>("assertion", token),  
                new KeyValuePair<string, string>("grant_type", "urn:ietf:params:oauth:grant-type:jwt-bearer")  
            });  

            var response = await httpClient.SendAsync(request);  
            var jsonResponse = await response.Content.ReadAsStringAsync();  

            if (response.IsSuccessStatusCode)  
            {  
                dynamic result = JsonConvert.DeserializeObject(jsonResponse);  
                return result.access_token;  
            }  
            else  
            {  
                throw new Exception($"Error retrieving access token: {jsonResponse}");  
            }  
        }  
    }  

    // Создание JWT для аутентификации  
    static string CreateJwt(ServiceAccountCredential credential)  
    {  
        var header = new  
        {  
            alg = "RS256",  
            typ = "JWT"  
        };  

        var claims = new  
        {  
            iss = credential.ClientEmail,  
            scope = "https://www.googleapis.com/auth/spreadsheets",  
            aud = "https://oauth2.googleapis.com/token",  
            exp = (int)(DateTime.UtcNow.AddMinutes(60) - new DateTime(1970, 1, 1)).TotalSeconds,  
            iat = (int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds  
        };  

        // Кодирование заголовка и полезной нагрузки в Base64Url  
        //string headerEncoded = Convert.ToBase64String(Encoding.UTF8.GetBytes(JsonConvert.SerializeObject(header)));  
        //string claimsEncoded = Convert.ToBase64String(Encoding.UTF8.GetBytes(JsonConvert.SerializeObject(claims)));  
        // Кодирование заголовка и полезной нагрузки в Base64Url  
        string headerEncoded = Base64UrlEncode(JsonConvert.SerializeObject(header));  
        string claimsEncoded = Base64UrlEncode(JsonConvert.SerializeObject(claims));  

        // Создание подписи на основании заголовка и полезной нагрузки 
        string signature = CreateSignature(headerEncoded + "." + claimsEncoded, credential.PrivateKey);  

        // Возвращаем полный JWT 
        return headerEncoded + "." + claimsEncoded + "." + signature;   
    }  

    // Вспомогательный метод для кодирования в Base64, с учетом различий в форматах  
    static string Base64UrlEncode(string input)  
    {  
        var bytes = Encoding.UTF8.GetBytes(input);  
        var base64 = Convert.ToBase64String(bytes);  
        return base64.Replace("=", "").Replace("+", "-").Replace("/", "_");  
    } 

    // Подпись JWT  
    static string CreateSignature(string message, string privateKey)  
    {  
        if (string.IsNullOrWhiteSpace(message))  
        {  
            throw new ArgumentException("Message cannot be null or empty.", nameof(message));  
        }  

        if (string.IsNullOrWhiteSpace(privateKey))  
        {  
            throw new ArgumentException("Private key cannot be null or empty.", nameof(privateKey));  
        }  

        using (var rsa = RSA.Create())  
        {  
            // Импортируем закрытый ключ из PEM  
            rsa.ImportFromPem(privateKey.ToCharArray());  
            
            // Подписываем данные  
            var signature = rsa.SignData(Encoding.UTF8.GetBytes(message), HashAlgorithmName.SHA256, RSASignaturePadding.Pkcs1);  
            
            // Возвращаем подпись в формате Base64  
            return Convert.ToBase64String(signature);  
        } 
    }  

    // Проверка доступа к Google-таблице  
    static async Task<bool> CheckSheetAccess(string spreadsheetId, string range, string accessToken)  
    {  
        // Проверка входных параметров  
        if (string.IsNullOrWhiteSpace(spreadsheetId))  
        {  
            throw new ArgumentException("Spreadsheet ID cannot be null or empty.", nameof(spreadsheetId));  
        }  

        if (string.IsNullOrWhiteSpace(range))  
        {  
            throw new ArgumentException("Range cannot be null or empty.", nameof(range));  
        }  

        if (string.IsNullOrWhiteSpace(accessToken))  
        {  
            throw new ArgumentException("Access token cannot be null or empty.", nameof(accessToken));  
        }  

        // Использование общего экземпляра httpClient  
        httpClient.DefaultRequestHeaders.Authorization = new AuthenticationHeaderValue("Bearer", accessToken);  
        var url = $"https://sheets.googleapis.com/v4/spreadsheets/{spreadsheetId}/values/{range}";  

        try  
        {  
            HttpResponseMessage response = await httpClient.GetAsync(url);  

            // Проверка на успех  
            if (response.IsSuccessStatusCode)  
            {  
                return true;  
            }  
            else  
            {  
                string responseBody = await response.Content.ReadAsStringAsync();  
                Console.WriteLine($"Error accessing sheet: {response.StatusCode} - {responseBody}");  
                return false;  
            }  
        }  
        catch (HttpRequestException e)  
        {  
            // Обработка исключения сети  
            Console.WriteLine($"Request error: {e.Message}");  
            return false;  
        } 
    }  

    // Метод авторизации в ADVANTA для выполнеия запросов web-api  
    private static async Task<string> AuthorizeInAdvanta()   
    {  
        Console.WriteLine("Начинаю авторизацию в ADVANTA"); 
        
        // Переменная, для возврата результата работы метода
        string result = ""; 
        
        using (var httpClient = new HttpClient())  
        {  
            // Формируем данные для запроса  
            var loginData = new  
            {  
                Login = a2login,  
                Password = a2password  
            };  

            var content = new StringContent(JsonConvert.SerializeObject(loginData), Encoding.UTF8, "application/json");  

            // Выполняем POST-запрос  
            var response = await httpClient.PostAsync($"{a2serverUrl}/api/auth/login", content);  

            if (response.IsSuccessStatusCode)  
            {  
                // Извлекаем куки из заголовка Set-Cookie  
                IEnumerable<string>? setCookieHeaders;  
                if (response.Headers.TryGetValues("Set-Cookie", out setCookieHeaders))  
                {  
                    // Берем первое значение куки и отсекаем дополнительную информацию  
                    var fullSetCookie = setCookieHeaders.FirstOrDefault()?.Trim();  
                    if (fullSetCookie != null)  
                    {  
                        // Берём только значение самого куки, игнорируя дополнительные параметры для браузера
                        var semicolonIndex = fullSetCookie.IndexOf(';');  
                        var cookie = semicolonIndex > 0 ? fullSetCookie.Substring(0, semicolonIndex) : fullSetCookie;
                        Console.WriteLine("Куки успешно извлечены: " + cookie); 
                        result = cookie;
                    }  
                }  
                else  
                {  
                    Console.WriteLine("Куки не найдены в заголовке ответа."); 
                    result = string.Empty; 
                }  
            }  
            else  
            {  
                Console.WriteLine("Ошибка авторизации: " + response.ReasonPhrase);  
                result = string.Empty;
            }  
        }  

        return result;
    }

    // Метод получает список объектов из ADVANTA, через LINQ-запрос 
    public static async Task<List<AdvantaObject>> GetObjectsList(string cookie)  
    {  
        Console.WriteLine("Начинаю вызов LINQ-запроса в ADVANTA"); 
        
        var advantaObjects = new List<AdvantaObject>(); // Инициализируем пустой список

        using (var httpClient = new HttpClient())  
        {  
            // Устанавливаем хост  
            httpClient.BaseAddress = new Uri(a2serverUrl);  
            
            // Формируем данные для запроса  
            var queryData = new  
            {  
                DataSourceKey = a2dataSourceKey,  
                PageSize = a2numOfObjects  
            };  

            var content = new StringContent(JsonConvert.SerializeObject(queryData), Encoding.UTF8, "application/json");  

            // Добавляем заголовок куки  
            if (!string.IsNullOrEmpty(cookie))  
            {  
                httpClient.DefaultRequestHeaders.Add("Cookie", cookie);  
            }  

            // Выполняем POST-запрос на указанный URL  
            var response = await httpClient.PostAsync("/api/queries/get", content);  

            if (response.IsSuccessStatusCode)  
            {  
                var responseData = await response.Content.ReadAsStringAsync();   

                // Десериализуем ответ в список объектов  
                var deserializedObjects = JsonConvert.DeserializeObject<List<AdvantaObject>>(responseData);  
                if (deserializedObjects != null)  
                {  
                    advantaObjects = deserializedObjects; // Присваиваем либо восстановленный список, либо пустой  
                } 
            }  
            else  
            {  
                Console.WriteLine("Ошибка получения списка объектов: " + response.ReasonPhrase);  
            }  
        }  

        return advantaObjects; // Всегда возвращаем ненулевой список
    }

    // Метод очищает указанную Google-таблицу  
    public static async Task ClearSheet(string spreadsheetId, string sheetName, string accessToken)  
    {  
        if (string.IsNullOrWhiteSpace(spreadsheetId))  
        {  
            throw new ArgumentException("Идентификатор таблицы не может быть пуст или иметь значение null.", nameof(spreadsheetId));  
        }  

        if (string.IsNullOrWhiteSpace(sheetName))  
        {  
            throw new ArgumentException("Имя листа не может быть пусто или иметь значение null.", nameof(sheetName));  
        }  

        if (string.IsNullOrWhiteSpace(accessToken))  
        {  
            throw new ArgumentException("Токен доступа не может быть пуст или иметь значение null.", nameof(accessToken));  
        }  

        Console.WriteLine("Начинаю очистку Google-таблицы");  

        // Устанавливаем заголовки запроса, включая токен доступа  
        httpClient.DefaultRequestHeaders.Authorization = new AuthenticationHeaderValue("Bearer", accessToken);  

        // Определяем диапазон, который хотим очистить  
        string range = $"{sheetName}!A1:Z";  

        // Создаем JSON-объект для запроса очистки  
        var clearRequest = new  
        {  
            range = range  
        };  

        // Отправляем POST-запрос на очистку значений в указанном диапазоне  
        string url = $"https://sheets.googleapis.com/v4/spreadsheets/{spreadsheetId}/values/{range}:clear";  
        
        try  
        {  
            HttpResponseMessage response = await httpClient.PostAsync(url, new StringContent(JsonConvert.SerializeObject(clearRequest), Encoding.UTF8, "application/json"));  

            // Проверяем статус ответа  
            if (response.IsSuccessStatusCode)  
            {  
                Console.WriteLine("Google-таблица успешно очищена.");  
            }  
            else  
            {  
                string responseBody = await response.Content.ReadAsStringAsync();  
                Console.WriteLine($"Ошибка при очистке Google-таблицы: {response.StatusCode} - {responseBody}");  
            }  
        }  
        catch (HttpRequestException e)  
        {  
            Console.WriteLine($"Ошибка запроса: {e.Message}");  
        }  
    }  

    // Записывает объекты в ячейки Google-таблицы
    static async Task WriteAdvantaObjectsToSheet(string spreadsheetId, string sheetName, List<AdvantaObject> advantaObjects, string accessToken)   
    {  
        // Проверка входных параметров  
        if (string.IsNullOrWhiteSpace(spreadsheetId))  
            throw new ArgumentException("ID таблицы не может быть пусто или null.", nameof(spreadsheetId));  

        if (string.IsNullOrWhiteSpace(sheetName))  
            throw new ArgumentException("Имя листа не может быть пусто или null.", nameof(sheetName));  

        if (advantaObjects == null || advantaObjects.Count == 0)  
            throw new ArgumentException("Список объектов ADVANTA не может быть пуст.", nameof(advantaObjects));   

        if (string.IsNullOrWhiteSpace(accessToken))  
            throw new ArgumentException("Access token не может быть пуст или null.", nameof(accessToken));  

        Console.WriteLine("Начинаю запись строк в Google-таблицу.");   

        // Подготавливаем данные для записи  
        var values = new List<IList<object>>  
        {  
            new List<object> { "Id", "Name", "PercentComplete", "Status", "Priority" } // Заголовки  
        };  

        foreach (var obj in advantaObjects)  
        {  
            values.Add(new List<object> { obj.Id, obj.Name, obj.PercentComplete, obj.Status, obj.Priority });  
        }  

        // Формируем тело запроса  
        var body = new  
        {  
            range = $"{sheetName}!A1", // Диапазон для записи  
            values = values  
        };  

        // Формируем URL для запроса  
        string url = $"https://sheets.googleapis.com/v4/spreadsheets/{spreadsheetId}/values/{sheetName}!A1:append?valueInputOption=USER_ENTERED";  

        // Устанавливаем заголовок авторизации  
        httpClient.DefaultRequestHeaders.Authorization = new AuthenticationHeaderValue("Bearer", accessToken);  

        try  
        {  
            // Отправляем PUT-запрос с данными  
            HttpResponseMessage response = await httpClient.PostAsync(url, new StringContent(JsonConvert.SerializeObject(body), Encoding.UTF8, "application/json"));  

            if (response.IsSuccessStatusCode)  
            {  
                Console.WriteLine("Данные успешно записаны в таблицу Google Sheets.");  
            }  
            else  
            {  
                string responseBody = await response.Content.ReadAsStringAsync();  
                Console.WriteLine($"Ошибка при записи данных в Google Sheets: {response.StatusCode} - {responseBody}");  
            }  
        }  
        catch (HttpRequestException ex)  
        {  
            Console.WriteLine("Ошибка запроса при записи данных в Google Sheets: " + ex.Message);  
        }  
    } 

    // Класс для представления учетных данных сервисного аккаунта      
    private class ServiceAccountCredential  
    {  
        [JsonProperty("client_email")]  
        public string ClientEmail { get; set; }  

        [JsonProperty("private_key")]  
        public string PrivateKey { get; set; }  
    } 
}